#!/usr/bin/env bash
# @(#) Generate a channel spec file based on user input.

errmsg()
{
  echo >&2 "FATAL-ERROR: $@"; exit 1
}

errdlg()
{
    dialog --title 'CRITICAL ERROR' --infobox "$@" 30 50; exit 1
}

complain()
{
    dialog --title 'ERROR' --msgbox "$@" 30 50
}


#
# Variables:
#

APP=$0
SCRIPT=$(basename $0)

hostnm=$(hostname)
: ${MAIL_AGENT:='mailx'}

OSNAME=$(uname -s)
case "${OSNAME:?}" in
    'Linux')    dialog_pkg=dialog
                CONFDIR='/etc'
                grep -q adm /etc/group && AGROUP=adm
                ;;
    'FreeBSD')  dialog_pkg=cdialog
                CONFDIR='/usr/local/etc'
                AGROUP=wheel
                ;;
     *) errmsg "Unsupported OS: ${OSNAME}" ;;
esac
type dialog 2>/dev/null >&2 || errmsg "${dialog_pkg} (>= 1.3) must be available"

LOHIMARK_INFO="\
Low:    size to keep before doing any archiving;\n\
High:   safe limit before a disk-usage alert;"


#
# Functions:
#

ask_admin_email()
{
    local info_txt="Please specify email for the channel administrator.\
 It will be used to send alarms, notices and other important information \
 regarding system performance.\n\n\
 Type 'none' if you want NO email notifications.\n\n\
 Administrator email:"

    local answer=/tmp/cha.aemail.$$
    local lrc=

    dialog --inputbox "${info_txt}" 16 60 ${cha_ADMIN_EMAIL} 2>$answer
    lrc=$?
    [ $lrc -eq 0 ] && cha_ADMIN_EMAIL=$(cat $answer)
    rm -f $answer

    return $lrc
}


setup_menu()
{
    local answer=/tmp/cha.smenu.$$
    local lrc=
    local dsize='-'
    local url=${cha_URL:--}

    [ -n "${cha_LMARK_DISKSIZE}" ] && \
        dsize="${cha_LMARK_DISKSIZE:--}/${cha_HMARK_DISKSIZE:--} Mb"

    : ${cha_ADMIN_EMAIL:=${GXA_ADMIN_EMAIL}}
    : ${cha_ROOT:=${GXA_ROOT}/channel}


    local create_status='(NEED MORE DATA)'
    if can_make_spec; then
        create_status='(READY)'
    fi

    if [ "${#url}" -gt '20' ]; then
        url="${url:0:17}..."
    fi

    dialog --menu 'Channel setup' 20 78 18                          \
        Name        "Channel ID [${cha_ID:--}]"                     \
        URL         "Source URL [${url}]"                           \
        Capacity    "Data storage capacity [${cha_CAPACITY:--} sec]"   \
        Duration    "Data-segment duration [${cha_DURATION:--} sec]"   \
        DiskSize    "Low/High-mark disk size quotas [${dsize}]"     \
        Data        "Root directory [${cha_ROOT:--}]"               \
        Shards      "Data shards [${cha_SHARDS:-disabled}]"         \
        TimeFmt     "Data-folder time format [${cha_GRAN_MASK}]"    \
        Email       "Administrator's email [${cha_ADMIN_EMAIL:--}]" \
        Create      "Create channel ${create_status}"               \
        Exit        "Quit the setup"                                \
        2>${answer}; lrc=$?

    [ $lrc -eq 0 ] && CHOICE=$(cat $answer)

    rm -f $answer
    return $lrc
}


get_cha_id()
{
    local descr="Please provide unique ID for the channel.\
 No spaces. Alpha-numerics and dashes only, please.\n\n
 Channel ID:"

    local answer=/tmp/cha.id.$$
    local lrc=

    dialog --inputbox "$descr" 12 70 2>${answer}; lrc=$?
    [ "$lrc" -eq 0 ] && cha_ID=$(cat $answer | tr -d '\t ')

    rm -f ${answer}
    return $lrc
}


get_src_url()
{
    local descr='FFmpeg-style URL for channel source stream:'
    local answer=/tmp/cha.url.$$
    local lrc=
    local fsize=

    while true; do
        dialog --inputbox "$descr" 8 70 "${cha_URL}" 2>${answer}; lrc=$?
        [ "$lrc" -eq "0" ] || break

        cha_URL=$(cat $answer)
        if [ -z "${cha_URL}" ]; then
            lrc=1; break
        fi

        gxprobe 10 "${cha_URL}" && break
    done

    rm -f ${answer}
    return $lrc
}


get_fsize()
{
    case "${OSNAME}" in
        'Linux')    stat -c '%s' ${1:?} ;;
        'FreeBSD')  stat -f '%z' ${1:?} ;;
        *) errmsg "Unsupported OS: ${OSNAME}" ;;
    esac
}


gxprobe()
{
    local nsec=${1:?}
    local url=${2:?}
    local cdir=/tmp/${cha_ID:?}
    local glog=${cdir}/gxseg.log
    local gtxt=${cdir}/gxseg.out
    local gxuopt=
    local gpid=
    local lrc=1

    rm -fr ${cdir}
    PROBE_NSEC=; PROBE_FSIZE=

    mkdir -m 775 -p ${cdir}

    if [ '0' -eq "${EUID}" ]; then
        gxuopt="-c ${GXA_USER:?}"

        chown ${GXA_USER:?} ${cdir}
        [ -n "${AGROUP}" ] && chgrp ${AGROUP} ${cdir}
    fi

    gxseg -kq -A 5 -i "${url:?}" -S /tmp -N ${cha_ID} -U ${nsec} ${gxuopt} >${gtxt} 2>${glog} &
    gpid=$!

    dialog --title 'Probing' --pause "${url}" 8 75 $(($nsec + 15)); lrc=$?
    kill $gpid 2>/dev/null
    wait $gpid
    if [ '0' != "${lrc}" ]; then
        rm -fr ${cdir}
        return $lrc
    fi

    if [ -s "${gtxt}" ]; then
        PROBE_FSIZE=$(grep '^SEGMENT:' ${gtxt} | sed -n '1p' | cut -d' ' -f 5)
        if [ -n "${PROBE_FSIZE}" ]; then
            PROBE_NSEC=$nsec
        else
            lrc=1
        fi
    else
        rm -f ${gtxt}
        lrc=2
    fi

    if [ "$lrc" != '0' ]; then
        dialog --ok-label Accept --title "Error probing channel" --editbox "$glog" 30 70
        lrc=$?
    fi

    # Leave data behind, for troubleshooting.
    [ '0' = "$lrc" ] && rm -fr ${cdir}

    return $lrc
}


probe_hilo()
{
    local lrc=
    local MB=$((1024 * 1024))

    if [ -z "${PROBE_FSIZE}" ]; then
        local descr="${LOHIMARK_INFO:?}\n\nMeasure channel speed at [${cha_URL}]?"
        if dialog --title 'Data Lo/Hi disk quotas' --yesno "$descr" 15 70; then
            gxprobe 10 ${cha_URL:?} || return 1
        fi
    fi

    if [ -n "${PROBE_FSIZE}" -a -n "${PROBE_NSEC}" ]; then
        local kbps=$(($PROBE_FSIZE / $PROBE_NSEC / 1024))

        local lmark=$((${kbps} * ${cha_CAPACITY:?} * 3 / 4 ))
        local hmark=$((${kbps} * ${cha_CAPACITY:?} * 4 / 3 ))

        lmark=$(($lmark / 1024))
        hmark=$(($hmark / 1024 + 1))

        local info="\n${LOHIMARK_INFO}\n\n\
Sample size:         : ${PROBE_FSIZE} bytes\n\
Probe duration       : ${PROBE_NSEC} sec\n\
Speed                : ${kbps} KB/sec \n\
Storage capacity     : ${cha_CAPACITY} sec\n\n\
Low/High watermarks  : ${lmark}/${hmark} MB\n\n\
Accept these settings or set manually?"

        if dialog --title 'Estimated quotas:' --yes-label 'Accept' --no-label 'Edit' \
                --yesno "${info}" 30 60; then
            cha_LMARK_DISKSIZE=${lmark}
            cha_HMARK_DISKSIZE=${hmark}
            return 0
        fi
    fi

    return 1
}


get_root_dir()
{
    local lrc=
    local answer=/tmp/cha.droot.$$
    dialog --title 'Channel root' --dselect ${cha_ROOT} 10 40 2>${answer}; lrc=$?
    [ "$lrc" -eq 0 ] && cha_ROOT=$(cat $answer)

    rm -f ${answer}
    return $lrc
}


get_dcapacity()
{
    local descr="How much data (in minutes) should be stored?"
    local answer=/tmp/cha.dcap.$$

    local minutes=$((${cha_CAPACITY:-0} / 60))
    local lrc=

    dialog --inputbox "${descr}" 10 50 ${minutes} 2>${answer}; lrc=$?
    if [ 0 -eq "$lrc" ]; then
        minutes=$(cat $answer)
        cha_CAPACITY=$((${minutes} * 60))
    fi

    rm -f ${answer}
    return $lrc
}


get_duration()
{
    local descr="Segment duration (seconds):"
    local lrc=
    local answer=/tmp/cha.dur.$$

    dialog --inputbox "${descr}" 8 50 ${cha_DURATION} 2>${answer}; lrc=$?
    [ 0 -eq "$lrc" ] && cha_DURATION=$(cat ${answer})

    rm -f ${answer}
    return $lrc
}


check_shards()
{
    local old_ifs=$IFS
    local nshards=${1:?}

    IFS=:; for shard in ${nshards}; do
        sdir=${cha_ROOT}/${shard}
        if [ ! -d "${sdir}" ]; then
            complain "Missing shard directory [${sdir}]"
            return 1
        fi
    done
    return 0
}


get_shards()
{
    local lrc=

    local descr="Paths to (sharded) subdirectories. Given disk1:disk2 as shards, \
channel data balances between ${cha_ROOT}/disk1/${cha_ID} and \
${cha_ROOT}/disk2/${cha_ID}.\n\
Shard directories MUST be created in advance, setup will NOT set them up.\n\
Please separate shards by colon.\n\n\
Enter shards:"
    local answer=/tmp/cha.shards.$$

    dialog --inputbox "${descr}" 18 60 "${cha_SHARDS}" 2>${answer}; lrc=$?
    if [ 0 -eq "$lrc" ]; then
        local shards=$(cat ${answer})
        if check_shards $shards; then
            cha_SHARDS=$shards
        else
            lrc=1
        fi
    fi

    rm -f ${answer}
    return $lrc
}


get_time_fmt()
{
    local desc="Time format mask (see strftime(3)) for data \
sub-directories."
    local lrc=
    local answer=/tmp/cha.tmfmt.$$

    dialog --inputbox "${descr}" 12 60 ${cha_GRAN_MASK} 2>${answer}; lrc=$?
    [ 0 -eq "$lrc" ] && cha_GRAN_MASK=$(cat ${answer})

    rm -f ${answer}
    return $lrc
}


get_hilo_disk()
{
    local lrc=
    local answer=/tmp/cha.hilo.$$
    local descr="${LOHIMARK_INFO:?}\n\n\
 Enter low/high values in MB (separated by slash):"

    local hilo=
    [ -n "${cha_LMARK_DISKSIZE}" ] && \
        hilo="${cha_LMARK_DISKSIZE}/${cha_HMARK_DISKSIZE}"

    dialog --inputbox "${descr}" 20 60 "${hilo}" 2>${answer}; lrc=$?
    if [ "0" -eq "$lrc" ]; then
        cha_LMARK_DISKSIZE=$(cut -d'/' -f1 $answer)
        cha_HMARK_DISKSIZE=$(cut -d'/' -f2 $answer)
    fi

    rm -f ${answer}
    return $lrc
}


get_storage_limits()
{
    [ -n "${cha_URL}" ] && probe_hilo
    [ -n "${cha_LMARK_DISKSIZE}" ] || get_hilo_disk
}


setup_runtime_env()
{
    if echo ${cha_ID:-sample} | grep -qi sample ; then
        cha_ID=; cha_URL=
    fi

    cha_PFX=
    : ${GXA_USER:=gigaplus}
}


save_channel_spec()
{
    local specfile=${1:?}
    local lrc=
    local errtxt=/tmp/err.$$

    cat >${specfile} 2>$errtxt <<_EOT
#!/bin/sh
# @(#) channel spec for Gigapxy VSM

# Clients must reference channel using this unique ID.
cha_ID="${cha_ID:?}"

# Path to gxpm (domain) socket.
cha_GPM="/opt/gxa/run/gxpm-S.sock"

# ffmpeg-style URL for channel source.
cha_URL="${cha_URL:?}"

# Path to data root (channel name will be added to this path).
cha_ROOT="/opt/gxa/channel"

# Channel-specific prefix to playlist items.
#
# NB: if leaving commented out, do NOT forget to
#     set gpm.item_url_prefix setting in gigaplus.conf.
#
# cha_PFX="http://your.host:8181/channel/"

# Paths to (sharded) subdirectories.
# Shards must be specified as a colon-separated list.
# Full path would become: $cha_ROOT/shardN/$cha_ID.
# Example: vol1:vol2:vol3
#
cha_SHARDS="${cha_SHARDS}"

# How much data in seconds should we keep?
# Segments will be rotated according to this limit.
#
# Example: 259200 sec = 3 days
#
cha_CAPACITY=${cha_CAPACITY}

# Duration of a segment (in seconds).
# (Optional, default=5).
#
cha_DURATION=${cha_DURATION}

# Allow N critical errors w/o quitting the app.
# Irregularities in input cause libav to produce "critical errors"
# that require either an exit(3) of gxseg(1) or an internal
# restart of the app. N restarts will be performed before exit(3).
#
# NB: set to -1 to always re-start and never exit(3). Restarts
#     are known cause memory leaks (thus the fewer the better),
#     but are *fast* compared to app exits.
#
cha_MAX_CERR=10

# Transcode using these (ffmpeg) options.
# IMPORTANT: do test, as in: ffmpeg -i http://source/channel ${cha_TRANSOPT} -f mpegts out.ts
#
# cha_TRANSOPT="-map 0:0 -map 0:1 -map 0:2 -map 0:3 -vcodec copy -c:a aac -c:s copy"

# Roll over (to a backup instance of gxseg) every N seconds.
# NB: this is done to avoid libav-specific issue with PTS inconsistencies
# and stalling of data streams.
#
# NB: the preferred way is to use force-rollover.sh script, scheduling
# roll-overs in low-activity hours via cron(8).
#
# cha_ROLLOVER=3600

# Root for segments generated by backup gxseg instance before rollover.
#
cha_FALSE_ROOT=/opt/gxa/bufd

# Timestamp mask as per strftime(3) for segment subdirs.
cha_GRAN_MASK="${cha_GRAN_MASK}"

# Number of items to keep in the manifest.
# Unless set here, this value is automatically calculated as cha_CAPACITY / cha_DURATION + 10.
# If set to 0, the value is NOT set and no manifest is going to be kept.
#
# cha_MITEMS=50

# Channel size in MB (total size of all files) at full load
# (after cha_CAPACITY seconds passed).
#
# Note: hos won't archive segments until this size is reached.
#
cha_LMARK_DISKSIZE=${cha_LMARK_DISKSIZE}

# Maximum allowable size (in MB) for a channel.
# An alarm will be triggered if this size is exceeded.
#
cha_HMARK_DISKSIZE=${cha_HMARK_DISKSIZE}

# Notify administrator (by this email) on crucial events.
cha_ADMIN_EMAIL=${cha_ADMIN_EMAIL}

# vim: syntax=sh
# __EOF__

_EOT

    lrc=$?
    if [ "${lrc}" -ne '0' ]; then
        dialog --title 'Error saving channel spec' --editbox "${errtxt}" 20 65
        return 1
    else
        rm ${errtxt}
    fi

    return 0
}


mk_channel_spec()
{
    local tmp_specfile=/tmp/${cha_ID:?}.$$

    save_channel_spec $tmp_specfile

    if dialog --title 'Confirm/edit channel spec' --editbox $tmp_specfile 50 80
    then
        SPECFILE=/tmp/${cha_ID:?}.spec
        mv ${tmp_specfile} ${SPECFILE} || return $?
    else
        rm -f ${tmp_specfile}
    fi

    return 0
}


can_make_spec()
{
    ERRMSG=
    [ -n "${cha_ID}" -a -n "${cha_URL}" ] \
                                || ERRMSG="(x) Channel ID and URL must be set.\n${ERRMSG}"
    [ -n "${cha_ADMIN_EMAIL}" ] || ERRMSG="(x) Admin email must be set\n${ERRMSG}"
    [ -n "${cha_DURATION}" -a -n "${cha_CAPACITY}" ] \
                                || ERRMSG="(x) Duration and Capacity must be set\n${ERRMSG}"
    [ -n "${cha_LMARK_DISKSIZE}" -a -n "${cha_LMARK_DISKSIZE}" ] \
                                || ERRMSG="(x) Low/High-mark disk quotas must be set\n${ERRMSG}"

    [ -n "${ERRMSG}" ] && return 1
    return 0
}


create_channel()
{
    local height=10

    echo "Channel: [${cha_ID}]" >${STATUS}

    if ! can_make_spec; then
        dialog --title 'Cannot make channel spec' --msgbox "${ERRMSG}" 10 50
        return 1
    fi

    mk_channel_spec || return 1
    [ -f "${SPECFILE}" ] || errdlg "Missing ${SPECFILE}"
    . ${SPECFILE}

    local ddir="${cha_ROOT:?}/${cha_ID:?}"
    local susp=${ddir}/SUSPEND

    if [ ! -d "$ddir" ]; then
        mkdir -m 775 ${ddir} || return $?
        :> ${susp}

        if [ "$EUID" != "${GXA_USER:?}" ]; then
            chown  ${GXA_USER:?} ${ddir} ${susp} || return $?
        fi
        [ -n "${AGROUP}" ] && chgrp ${AGROUP} ${ddir} ${susp}

        echo "Directory: ${ddir}" >>${STATUS}
    fi

    if [ -n "${cha_SHARDS}" ]; then
        local old_ifs=$IFS
        local shard=
        local i=0
        IFS=:; for shard in ${cha_SHARDS}; do
            let i=i+1
            if [ -d "${cha_ROOT}/${shard}" ]; then
                sdir=${cha_ROOT}/${shard}/${cha_ID}
                if [ ! -d "${sdir}" ]; then
                    mkdir -m 775 ${sdir} || return $?
                    [ "$EUID" = "${GXA_USER}" ] || chown ${GXA_USER} ${sdir}
                    [ -n "${agroup}" ] && chgrp ${agroup} ${sdir}

                    echo "shard: ${sdir}" >>${STATUS}
                    let height+=1
                fi
            else
                complain "Missing ${shard} shard [${cha_ROOT}/${shard}]"; return 1
            fi
        done; IFS=${old_ifs}

        echo "$i shards" >>${STATUS}
    fi

    [ -n "$DEST_SPEC" ] || DEST_SPEC=${GXA_ROOT:?}/etc/spec/$(basename ${SPECFILE:?})
    if [ -f "${DEST_SPEC}" ]; then
        dialog --title "Spec already exists, overwrite?" ${DEST_SPEC} $height 70 || return 1
    fi

    cp ${SPECFILE} ${DEST_SPEC} || errdlg "Failed to create ${DEST_SPEC}"
    if [ '0' -eq "${EUID}" ]; then
        chown ${GXA_USER:?} ${DEST_SPEC}
        [ -n "${AGROUP}" ] && chgrp ${AGROUP} ${DEST_SPEC}
    fi

    echo "Spec: ${DEST_SPEC}" >>${STATUS}
    rm ${SPECFILE}

    rm -f ${OLD_SPEC}

    local cre_status="$(cat $STATUS)"
    if dialog --title "Channel created" --msgbox "${cre_status}" 20 60; then
        rm ${STATUS}
        exit 0
    fi

    return 0
}


full_setup()
{
    local text="This setup will guide you through essential steps \
to configure a channel. Or you always go to the main menu and choose your steps."

    dialog --title Setup --yes-label 'Continue' --no-label 'Menu' --yesno "$text" 12 40 || return

    get_cha_id          || return $?
    get_src_url         || return $?
    get_dcapacity       || return $?
    get_storage_limits  || return $?

    if [ -z "${cha_DURATION}" ]; then
        get_duration  || return $?
    fi

    if [ -z "${cha_ROOT}" ]; then
        get_root_dir    || return $?
    fi

    if [ -z "${cha_ADMIN_EMAIL}" ]; then
        ask_admin_email
    fi
}


#
# Main
#

if [ $# -lt 1 ]; then
    echo >&2 "Usage: ${APP:?} [-menu] template-file [dest-spec]"
    exit 1
fi

if [ '-menu' = "$1" ]; then
    SKIP_FS=1; shift
fi

TMPL_SPEC=${1:?}
if [ ! -f "${TMPL_SPEC}" ]; then
    TMPL_HERE=$(basename ${TMPL_SPEC})
    if [ -f "${TMPL_HERE}" ]; then
        TMPL_SPEC="${TMPL_HERE}"
    else
        errmsg "Missing template [${TMPL_SPEC}]"
    fi
fi

. ${TMPL_SPEC}

ENVFILE=gigaplus.env
[ -f "${ENVFILE}" ] || ENVFILE="${CONFDIR}/${ENVFILE}"
if [ -f "${ENVFILE}" ]; then
    . ${ENVFILE}
else
    errmsg "Missing $(basename ${ENVFILE}) or $ENVFILE"
fi

DEST_SPEC=$2

PATH=${PATH}:.
SPECFILE=
OLD_SPEC=/tmp/unsaved.spec
STATUS=/tmp/status.$$


if ! type gxseg >/dev/null 2>&1; then
    errmsg "Missing gxseg module"
fi

setup_runtime_env

if [ -f "$OLD_SPEC" ]; then
    if dialog --title 'Unfinished spec found' --ok-label 'Load' --cancel-label 'Discard' \
            --editbox "${OLD_SPEC}" 20 65; then
        . ${OLD_SPEC}
        SKIP_FS=1
    fi

    rm ${OLD_SPEC}
fi

if [ -z "${cha_ADMIN_EMAIL}" ]; then
    cha_ADMIN_EMAIL=${GXA_ADMIN_EMAIL}
fi


[ -n "${SKIP_FS}" ] || full_setup
QUIT=; while [ -z "${QUIT}" ]; do
    if [ -n "${cha_ID}" -a -n "${cha_URL}" ]; then
        save_channel_spec ${OLD_SPEC}
    fi

    setup_menu || break

    case "${CHOICE}" in
        Name)       get_cha_id          ;;
        URL)        get_src_url         ;;
        Data)       get_root_dir        ;;
        Capacity)   get_dcapacity       ;;
        Duration)   get_duration        ;;
        Shards)     get_shards          ;;
        TimeFmt)    get_time_fmt        ;;
        DiskSize)   get_storage_limits  ;;
        Email)      ask_admin_email     ;;
        Create)     create_channel      ;;
        Exit)       break               ;;
    esac
done

rm -f ${SPECFILE} ${STATUS}

# __EOF__

