#!/usr/bin/env bash
# @(#) GigA+ channel management script.

: ${GXA_PFX:='/opt/gxa'}

RM=/bin/rm
MKDIR=/bin/mkdir
CAT=/bin/cat

OSNAME=$(uname -s)
case "${OSNAME:?}" in
    'Linux'|'FreeBSD') ;;
     *) errmsg "Unsupported OS: ${OSNAME}" ;;
esac

hostnm=$(hostname)

: ${MAIL_AGENT:='mailx'}

GPX_PIDFILE_DIR=${GXA_PFX:?}/run
GPX_LOG_DIR=${GXA_PFX:?}/log
GPX_CRASH_DIR=${GXA_PFX:?}/core

GXPM_LOG=${GPX_LOG_DIR:?}/gxpm.log

#
# Functions:
#

errmsg()
{
  echo >&2 "FATAL-ERROR: $@"; exit 1
}

warn()
{
    echo >&2 "WARNING: $@"
}

msg()
{
    echo >&2 $@
}

dbg()
{
    [ -n "${GXA_DEBUG}" ] && echo >&2 "DEBUG: $@"
}


get_fmt_uptime()
{
    local asec=${1:?}

    local dy=$(( ${asec} / 86400 ))
    local dyl=$(( ${asec} % 86400 ))
    local hr=$(( ${dyl} / 3600 ))
    local hrl=$(( ${dyl} % 3600 ))
    local mi=$(( ${hrl} / 60 ))
    local sec=$(( ${hrl} % 60  ))

    if [ '--print' = "$2" ]; then
        printf "%02dd %02dh:%02dm:%02ds\n" $dy $hr $mi $sec
    else
        printf "%02dd %02dh:%02dm:%02ds" $dy $hr $mi $sec
    fi
}

print_fmt_uptime()
{
    get_fmt_uptime $1 --print
}


mod_utime()
{
    case "${OSNAME:?}" in
        'Linux')    stat -c '%Y' ${1:?} ;;
        'FreeBSD')  stat -f '%m' ${1:?} ;;
        *) errmsg "Unsupported OS: ${OSNAME}" ;;
    esac
}


get_file_age()
{
    local fpath=${1:?}
    local last_mod_tm=$(mod_utime ${fpath})

    local now=$(date +"%s")
    local asec=$((${now:?} - ${last_mod_tm}))
    print_fmt_uptime $asec
}


lnx_can_suid()
{
    local mode=${1:?}
    local procf='/proc/sys/fs/suid_dumpable'
    case "$mode" in
        --check) cat $procf ;;
        --show) echo "as root: echo 1 >$procf" ;;
        *) errmsg "Invalid mode: $mode" ;;
    esac
}


bsd_can_suid()
{
    local mode=${1:?}
    local setting='kern.sugid_coredump'

    case "$mode" in
        --check) sysctl $setting | cut -d':' -f2 | tr -d ' ' ;;
        --show) echo "as root: sysctl ${setting}=1" ;;
        *) errmsg "Invalid mode: $mode" ;;
    esac
}


get_can_suid()
{
    local mode=${1:?}
    case "${OSNAME:?}" in
        'Linux')    lnx_can_suid ${mode} ;;
        'FreeBSD')  bsd_can_suid ${mode} ;;
    esac
}


get_core_dir()
{
    case "${OSNAME:?}" in
        'Linux')    dirname $(cat /proc/sys/kernel/core_pattern) ;;
        'FreeBSD')  dirname $(sysctl kern.corefile | cut -d':' -f2 | tr -d ' ') ;;
    esac
}


get_channel_size()
{
    local duopt=${1:?}
    local croot=${2:?}
    local ctag=${3:?}
    local shards=${4}

    local csize=$(du ${duopt} ${croot}/${ctag} | cut -f1)
    if [ -n "${shards}" ]; then
        for s in $(echo ${shards} | tr ':' ' '); do
            local shdir=${croot}/${s:?}/${ctag}
            local shsize=$(du ${duopt} ${shdir} | cut -f1)
            if [ -n "${shsize}" ]; then
                let csize+=${shsize}
            fi
        done
    fi

    echo ${csize}
}


email_alert()
{
    local subj="$1"
    local head="$2"
    local rc=

    [ -z "${can_email}" ] && return 1
    [ -n "${cha_ADMIN_EMAIL}" ] || return 1

    local letter=${GXA_ROOT:?}/log/alert-email.$$
    echo "$head" >${letter}

    local vlog="${cha_ROOT:?}/${cha_ID:?}/vsm.log"
    if [ -f "$vlog" ]; then
        echo "==== ${vlog} ====" >>${letter}
        tail -50 ${vlog} >>${letter}
        echo "=================" >>${letter}
    else
        echo "No ${vlog} found, listing ${cha_ROOT:?}/${cha_ID:?}:" >>${letter}
        ls -l ${cha_ROOT:?}/${cha_ID:?} >>${letter}
    fi

    cat ${letter} | ${MAIL_AGENT} -s "${subj}" ${cha_ADMIN_EMAIL:?}; rc=$?
    [ "0" = "$rc" ] && rm -f ${letter}
}


unset_spec_env()
{
    unset cha_ID cha_GPM cha_URL cha_ROOT cha_ROLLOVER
    unset cha_TRANSOPT cha_SHARDS
    unset cha_CAPACITY cha_DURATION cha_MAX_CERR
    unset cha_FALSE_ROOT cha_GRAN_MASK cha_MITEMS
    unset cha_LMARK_DISKSIZE cha_HMARK_DISKSIZE
    unset cha_HOS cha_PRB cha_ADMIN_EMAIL
}


cha_status()
{
    local specfile=${1:?}
    local alert=
    local online="OFF"

    [ -f "${specfile}" ] || errmsg "Invalid spec: ${specfile}"
    unset_spec_env
    . ${specfile} >/dev/null

    local cdir=${cha_ROOT:?}/${cha_ID:?}
    if [ ! -d "${cdir}" ]; then
        echo >&2 "(ERR) ${cha_ID:?} - [$cdir] not found"
        return 1
    fi

    local csize=$(get_channel_size -Lsm ${cha_ROOT:?} ${cha_ID:?} ${cha_SHARDS})
    if [ -n "${cha_HMARK_DISKSIZE}" ]; then
        if [ "$csize" -ge "${cha_HMARK_DISKSIZE}" ]; then
            alert=1; aux="${aux} *OVSZ*"
        fi
    fi

    local pidfile=$cdir/vsm.pid
    if [ -f "$pidfile" ]; then
        local vsmpid=$(cat $pidfile)
        if ps -p ${vsmpid} 2>/dev/null >&2; then
            online="ON"
            aux="uptime: $(get_file_age ${pidfile})"
        else
            local old_pid=$(cat ${pidfile} 2>/dev/null)
            if rm -f ${pidfile}; then
                aux="- stale pidfile (${old_pid}) removed; "
            fi
        fi
    fi

    [ -f $cdir/REPLICA ] && online="REPL"
    [ -f $cdir/SUSPEND ] && online="SUSP"

    if [ "OFF" = "${online}" ]; then
        alert=1
    fi

    [ -n "${MUST_ALERT}" ] || alert=

    if [ -n "${alert}" ]; then
        aux="${aux} (ALERT!)"
    fi

    local report="(${online}) ${cha_ID:?} [${csize} MB] ${aux}"
    echo >&2 "$report"; aux=

    if [ -n "${alert}" ]; then
        email_alert "channel alert for [${cha_ID}] on ${hostnm}" "${report}"
        alert=
    fi

    [ "ON" = "${online}" ] && return 0
    return 1
}

get_status()
{
    local tag=${1:?}
    local cha=

    echo >&2 "----- channels -----"
    if [ "all" = "$tag" ]; then
        find ${GXA_SPEC_DIR} -type f -name '*.spec' | \
            while read sname; do
                cha_status ${sname:?}
                cha=${cha_ID}
            done
    else
        for cha in `echo ${tag:?}`; do
            sname=${GXA_SPEC_DIR}/${cha:?}.spec
            if [ -f "${sname}" ]; then
                cha_status ${sname:?}
            else
                echo >&2 "No spec [${sname}] for ${cha}"
            fi
        done
    fi
    echo >&2 "--------------------"
}

upkeep_file()
{
    local cmd_file=${1:?}
    local dir=$(dirname ${cmd_file})
    local base=$(basename ${cmd_file})

    echo ${dir}/upkeep-${base}
}


check_pidfile()
{
    local hush=
    local uptime=

    pid=
    UPK_PID=

    if [ '--hush' = "$1" ]; then
        hush=1; shift
    fi

    local fname=$1
    local app=${2:-'process'}
    local run_msg=${3:-'running'}

    if [ -f "${fname}" ]; then
        uptime=$(($(date +'%s') - $(mod_utime ${fname})))
        pid=$(cat ${fname})
        [ -n "$pid" ] || return 1

        if ! ps -p ${pid} 2>/dev/null >&2; then
            [ -z "$hush" ] && echo "Stale pidfile [${fname}] found; removing."
            rm -f ${fname}
        else
            run_msg="${run_msg}; uptime: $(get_fmt_uptime ${uptime})"
        fi
    fi

    if [ ! -f "${fname}" ]; then
        [ -z "$hush" ] && echo "[${app}] is not running."
        return 2
    fi

    local upk_fname=$(upkeep_file $fname)
    dbg "${app} UPK: ${upk_fname}"

    if [ -f "${upk_fname}" ]; then
        UPK_PID=$(cat ${upk_fname})

        ps -p ${UPK_PID} 2>/dev/null >&2 || UPK_PID=
        if [ -n "${UPK_PID}" ]; then
            run_msg="${run_msg} | U[${UPK_PID}]"
        fi
    else
        dbg "No upkeep pid: ${upk_fname}"
    fi

    [ -z "$hush" ] && echo "${app} [${pid}] ${run_msg}"
    return 0
}

getoparam()
{
    local flag=${1:?}
    shift; while [ -n "$1" -a "$flag" != "$1" ]; do shift; done
    [ -n "$1" ] && echo $2
}


start_process()
{
    local cmd=${1:?}
    shift
    local options=$@
    local cmd_pidfile=$(getoparam -p $options)
    local lrc=

    [ -n "${cmd_pidfile}" ] || errmsg "-p pidfile must be defined"

    dbg "${cmd} ${options}"
    ${cmd} ${options}; lrc=$?
    [ '0' -eq "$lrc" ] || return $lrc

    if [ -n "${UPKEEP}" ]; then
        local cmd_logfile=$(getoparam -l $options)

        local upk_logfile=$(upkeep_file $cmd_logfile)
        local upk_pidfile=$(upkeep_file $cmd_pidfile)

        local uopt="-d -L $(dirname ${cmd_logfile}) -A ${GPX_CRASH_DIR} -p ${cmd_pidfile} -l ${upk_logfile}"
        uopt="${uopt} -F ${upk_pidfile} -U ${GXA_USER:?}"
        [ -n "$CORE_DIR" ] && uopt="${uopt} -c ${CORE_DIR}"

        sleep 0.3
        check_pidfile --hush ${cmd_pidfile} $(basename ${cmd}) || return 1

        dbg "${UPKEEP} $uopt -- ${cmd} ${options}"

        ${UPKEEP} $uopt -- ${cmd} ${options} >&2 &
        local upk_pid=$!
        sleep 0.3
        if kill -0 ${upk_pid} 2>/dev/null; then
            dbg "${UPKEEP}($upk_pid) running"
            disown ${upk_pid}
        fi
    fi
}


kill_process()
{
    local name=${1:?}
    local cmd_pid=${2:?}
    local upk_pid=$3

    if [ -n "${upk_pid}" ] && ps -p ${upk_pid} 2>/dev/null >&2; then
        dbg "notifying $name U[${upk_pid}]"
        if kill ${upk_pid}; then
            echo "${name}[${cmd_pid}] U[${upk_pid}] asked to stop"
            sleep 0.5
        else
            echo "${name}[${cmd_pid}] U[${upk_pid}] FAILED to stop"
        fi
    fi

    if ps -p ${cmd_pid} >/dev/null 2>&1; then
        dbg "${name}[${cmd_pid}] running (yet)"
        if kill ${cmd_pid}; then
            echo "${name} [${cmd_pid}] stopped"
        else
            echo "${name} [${cmd_pid}] FAILED to stop"
        fi
    else
        dbg "${name}[${cmd_pid}] down (already)"
    fi
}


ck_start_gxpm()
{

    if ! check_pidfile --hush ${GXPM_PIDFILE} 'gxpm'; then
        echo "Starting gxpm ..."
        start_process ${GPM:?} -q -p ${GXPM_PIDFILE} -l ${GXPM_LOG} ${GXPM_OPTIONS} || \
            errmsg "Failed to start GXPM"
    fi
}


ck_stop_gxpm()
{
    if check_pidfile --hush ${GXPM_PIDFILE} 'gxpm'; then
        kill_process 'gxpm' ${pid} ${UPK_PID}
    fi
}


ck_start_channel()
{
    local tag=${1:?}
    local cdir=${GXA_CHANNEL_ROOT:?}/${tag}
    local pidfile=${cdir:?}/vsm.pid

    if [ -f "$cdir/SUSPEND" ]; then
        echo >&2 "<skip> ${tag} (SUSP)"
        return 1
    fi
    if [ -f "$cdir/REPLICA" ]; then
        echo >&2 "<skip> ${tag} (REPL)"
        return 1
    fi

    check_pidfile --hush ${pidfile} ${tag} && return 0

    local logfile=${cdir:?}/vsm.log
    local specfile=${GXA_SPEC_DIR:?}/${tag}.spec

    sudo -u ${GXA_USER:-gigaplus} nohup ${VSM:?} -l ${logfile} ${specfile} >/tmp/nohup.out &

    local wpid=$!
    sleep 0.5
    if ! kill -0 $wpid 2>/dev/null; then
        echo >&2 "${tag}: FAILED to start (pid=$wpid)"
    else
        check_pidfile ${pidfile} ${tag}
        disown $wpid
    fi
}


purge_cha_pidfiles()
{
    local cdir=${1:?}
    local pfile=
    find ${cdir} -name '*.pid' | while read pfile; do
        if check_pidfile --hush ${pfile}; then
            rm ${pfile} && echo >&2 "Removed $pfile"
        fi
    done
}


ck_stop_channel()
{
    local rc=0
    local tag=${1:?}
    local cdir=${GXA_CHANNEL_ROOT:?}/${tag}
    local pidfile=${cdir:?}/vsm.pid

    if check_pidfile --hush ${pidfile} ${tag}; then
        if kill ${pid} 2>/dev/null; then
            echo >&2 "${tag} VSM stopped (pid=$pid)"

            purge_cha_pidfiles ${cdir:?}

            let stopped+=1
            dbg "st=${stopped}"
            rc=0
        fi
    else
        echo >&2 "${tag} <not-running>"
        rc=0
    fi

    if ls ${cdir:?}/*.pid 2>/dev/null; then
        echo >&2 "Some (${tag}) apps may be running, checking..."
        sleep 1
        for gpfile in $(ls -1 ${cdir:?}/*.pid 2>/dev/null); do
            if check_pidfile ${gpfile} ${tag}; then
                kill ${pid} || rc=1
                [ $rc -eq 0 ] && echo >&2 "STOPPED app($pid): $gpfile"
            fi
        done
    fi

    return ${rc}
}


ck_start_dwg()
{
    local inum=${1}     # can be empty
    local options=${2}  # can be empty

    local logfile="${GPX_LOG_DIR:?}/${app}${inum}.log"
    local pfile="${GXA_ROOT:?}/run/${app:?}${inum}.pid"

    if ! check_pidfile --hush ${pfile:?} ${DWG}; then
        echo "Starting ${DWG:?}#${inum}"
        start_process ${DWG} -q -p ${pfile:?} -l ${logfile:?} ${options} || \
            errmsg "Failed to start ${DWG}"
    fi

    check_pidfile ${pfile:?} ${DWG}
}


stop_dwags()
{
    local pfile=
    find ${GXA_ROOT:?}/run -type f -name 'dwg*.pid' | \
    while read pfile; do
        if check_pidfile --hush ${pfile} ${DWG}; then
            kill_process 'dwg' ${pid} ${UPK_PID}
        fi
    done
}


status_dwags()
{
    local pfile=

    find ${GXA_ROOT:?}/run -type f -name 'dwg*.pid' | \
    while read pfile; do
        check_pidfile "${pfile}" 'dwg'
    done
}


start_dwags()
{
    local count=${1:?}
    local app='dwg'

    local i; for (( i=1; i <= ${count}; i++ )); do
        local opt="${DWG_OPTIONS}"
        [ -n "${PIN_DWG}" ] && opt="${DWG_OPTIONS} --cpu $i"
        ck_start_dwg "$i" "${opt}" || break
    done
}


start_lb()
{
    local logfile=${GPX_LOG_DIR:?}/flb.log

    if ! check_pidfile --hush ${LB_PIDFILE:?} ${FLB:?}; then
        echo "Starting flb ..."
        start_process ${FLB} ${LB_OPTIONS} -p "${LB_PIDFILE:?}" -l ${logfile} ||
            errmsg "Failed to start flb (${FLB})"
    fi
}


stop_lb()
{
    [ -n "${LB_PIDFILE}" -a -f "${LB_PIDFILE}" ] || return 0

    if check_pidfile --hush ${LB_PIDFILE:?} 'flb'; then
        kill_process 'flb' ${pid} ${UPK_PID}
    fi
}


status_lb()
{
    check_pidfile ${LB_PIDFILE:?} 'flb'
}


start_channels()
{
    local tag=${1:?}
    local cha=
    local rc=$?
    local started=0

    if [ -z "${NOPM}" ]; then
        ck_start_gxpm || return 1
    fi

    if [ "all" = "$tag" ]; then
        find ${GXA_SPEC_DIR} -type f -name '*.spec' | \
            while read sname; do
                cha=$(basename $sname | sed 's/\.spec$//g')
                ck_start_channel ${cha:?}; rc=$?
                [ "$rc" -eq 0 ] && let started+=1
            done
    else
        for cha in `echo ${tag:?}`; do
            ck_start_channel ${cha:?}; rc=$?
            [ "$rc" -eq 0 ] && let started+=1
        done
    fi

    if [ "${started}" -gt 0 ]; then
        echo >&2 "${started} channel(s) started"
        echo >&2 "Pausing for [${CHA_PAUSE}] seconds.."
        sleep ${CHA_PAUSE:?}
    fi

    get_status ${tag}

    check_pidfile ${GXPM_PIDFILE} 'gxpm'
    [ -n "${USE_DWAGS}" ] && status_dwags
    [ -f "${LB_PIDFILE}" ] && status_lb
}


stop_channels()
{
    local tag=${1:?}
    local cha=
    local clist=/tmp/clist.$$

    stopped=0
    echo >&2 "Stopping [${tag}]"
    if [ "all" = "$tag" ]; then
        find ${GXA_SPEC_DIR} -type f -name '*.spec' >${clist}
        while read sname; do
                cha=$(basename $sname | sed 's/\.spec$//g')
                ck_stop_channel ${cha:?}
        done <${clist}; rm -f ${clist}
    else
        for cha in `echo ${tag:?}`; do ck_stop_channel ${cha:?}; done
    fi

    echo >&2 "${stopped} channels stopped."
    if [ "$stopped" -gt 0 ]; then
        echo >&2 "Pausing.." && sleep 1
        get_status ${tag}
    fi
}


shut_down()
{
    stop_channels all
    ck_stop_gxpm
    stop_dwags

    stop_lb
}


check_prerequisites()
{
    type ${GPM:?} >/dev/null || errmsg "Missing gxpm=${GPM}"

    if [ ! -d "${GPX_CONFIG_DIR}" ]; then
    errmsg "$0: config directory [${GPX_CONFIG_DIR}] does not exist."
    fi

    [ '1' = "$(get_can_suid --check)" ] || \
        errmsg "please ensure GigA+ modules CAN core-dump; use $(get_can_suid --show)"

    CORE_DIR=$(get_core_dir)
    if [ '.' = "${CORE_DIR}" ]; then
        warn "cores would go to the current directory, you might want to change that."
        CORE_DIR=
    else
    [ -d "${CORE_DIR}" ] || errmsg "Invalid core directory: ${CORE_DIR}"
    fi

    if [ -z "${SKIP_UPK}" ]; then
        if [ -z "${UPKEEP}" ]; then
            UPKEEP=upkeep
            if ! type ${UPKEEP} >/dev/null 2>&1; then
                [ -n "${GPX_BIN_DIR}" ] && UPKEEP="${GPX_BIN_DIR}/upkeep"
            fi
        fi

        type ${UPKEEP} >/dev/null || UPKEEP=
        if [ -n "${UPKEEP}" -a -z "${GXA_ADMIN_EMAIL}" ]; then
            warn "No GXA_ADMIN_EMAIL, upkeep will be disabled"
        fi

        [ -n "${UPKEEP}" ] || warn "Cannot use upkeep"
        dbg "upkeep: ${UPKEEP}"
    fi

    if [ ! -d "${GPX_LOG_DIR}" ]; then
        ${MKDIR} -p ${GPX_LOG_DIR} || errmsg "$0: failed to mkdir [${GPX_LOG_DIR}]"
    fi

    if [ -n "${USE_DWAGS}" ]; then
        type ${DWG:?} >/dev/null 2>&1 || errmsg "Missing dwg=${DWG}"

        if ! ls ${GXA_ROOT:?}/lbdb/sha1/*.sha1 >/dev/null 2>&1; then
            ${CAT:?} >&2 <<_EOT
----
In order to use ${DWG}, GigA+ must be configured for LB (load balancing).
This includes a number of steps, described in the documentation.

SEE: man gxa-lb-setup
----
_EOT
            errmsg 'Cannot start without proper LB setup.'
        fi
    fi
}


#
# Main
#

case ${OSNAME} in
    FreeBSD)
        [ -n "${GPX_CONFIG_DIR}" ] || GPX_CONFIG_DIR=/usr/local/etc
        [ -n "${GPX_BIN_DIR}" ] || GPX_BIN_DIR=/usr/local/bin
        ;;
    Linux)
        [ -n "${GPX_CONFIG_DIR}" ] || GPX_CONFIG_DIR=/etc
        [ -n "${GPX_BIN_DIR}" ] || GPX_BIN_DIR=/usr/bin
        ;;
    *) errmsg "${OSNAME} is not supported."
esac

if [ $# -lt 1 ]; then
cat <<_EOT >&2
Usage: $0 [--nopm] [--pause N] start|stop|status|shutdown [{channel}|all]

Options:
    --nopm              = do NOT attempt to start/stop gxpm.
    --pause             = wait N [5] seconds before checking on channels after start.
    --alarm             = generate alarm if channel is offline and/or oversized.
    --dwg[pin] {num}    = start/stop/status N dwg instances [pin to CPU cores].
    --no-upkeep         = disable upkeep (crash/restart & health monitor);
    --lb                = start/stop/status load balancer;
    --debug mask        = list app(s) in debug mode (lb:dwg:pm).
    -t                  = trace mode: see commands & trace info;
Note:
    'shutdown' stops all channels and modules.

GigA+ channel launch script.
This script is part of GigA+. Copyright 2016-2018 by Pavel Cherenkov
_EOT
    exit 1
fi


envfile="${GPX_CONFIG_DIR}/gigaplus.env"
[ -f "./gigaplus.env" ] && envfile="./gigaplus.env"

if [ -f "${envfile:?}" ]; then
    . ${envfile}
    : ${GXA_ROOT:?}
    : ${GXA_CHANNEL_ROOT:?}
else
    errmsg "Missing ${envfile}"
fi

: ${GXA_USER:=gigaplus}

[ '-' = "${GPX_BIN_DIR}" ] && GPX_BIN_DIR=

if [ -n "${GPX_BIN_DIR}" ]; then
    GPM=${GPX_BIN_DIR}/gxpm
    VSM=${GPX_BIN_DIR}/vsm
    DWG=${GPX_BIN_DIR}/dwg
    FLB=${GPX_BIN_DIR}/flb
else
    : ${GPM:=gxpm}
    : ${VSM:=vsm}
    : ${DWG:=dwg}
    : ${FLB:=flb}
fi

if type ${MAIL_AGENT:?} >/dev/null 2>&1; then
    can_email=1
fi

GXPM_PIDFILE="${GXA_ROOT:?}/run/gxpm.pid"
LB_PIDFILE="${GXA_ROOT:?}/run/flb.pid"

[ -n "${GXA_SPEC_DIR}" ] || GXA_SPEC_DIR="${GXA_ROOT:?}/etc"

while [ -n "$1" ]; do
    if [ '--nopm' = "$1" ]; then
        NOPM=1; shift
        continue
    fi

    if [ '--pause' = "$1" ]; then
        CHA_PAUSE=$2; shift 2
        continue
    else
        CHA_PAUSE=5
    fi

    if [ "--alert" = "$1" ]; then
        MUST_ALERT=1; shift
        continue
    fi

    if [ "--dwg" = "$1" ]; then
        USE_DWAGS=${2:?}; PIN_DWG=;
        if ! expr ${USE_DWAGS} + 1 >/dev/null 2>&1; then
            echo >&2 "Invalid --dwg parameter"; exit 1
        fi
        shift 2; continue
    fi

    if [ "--dwgpin" = "$1" ]; then
        USE_DWAGS=${2:?}; PIN_DWG=1;
        if ! expr ${USE_DWAGS} + 1 >/dev/null 2>&1; then
            echo >&2 "Invalid --dwgpin parameter"; exit 1
        fi
        shift 2; continue
    fi

    if [ "--lb" = "$1" ]; then
        USE_LB=1
        shift; continue
    fi

    if [ '--no-upkeep' = "$1" ]; then
        SKIP_UPK=1
        shift; continue
    fi

    if [ '--debug' = "$1" ]; then
        dmask=${2}
        case "$dmask" in
            *lb*)   LB_OPTIONS="-L debug ${LB_OPTIONS}";
                    echo >&2 "[lb-debug]"
                    ;;
        esac
        case "$dmask" in
            *dwg*)  DWG_OPTIONS="-L debug ${DWG_OPTIONS}"
                    echo >&2 "[dwg-debug]"
                    ;;
        esac
        case "$dmask" in
            *pm*)   GXPM_OPTIONS="-L debug ${GXPM_OPTIONS}"
                    echo "[gpxm-debug]"
                    ;;
        esac
        shift 2; continue
    fi

    if [ '-t' = "$1" ]; then
        GXA_DEBUG=1
        shift; continue
    fi

    break
done

cmd=${1}
if [ -z "$cmd" ]; then
    echo >&2 "Missing command (start, stop, etc.)"
    exit 1
fi

shift
cha_tag="$@"
[ -n "${cha_tag}" ] || cha_tag='all'

euid=$(id -u)
if [ "status" != "$cmd" ]; then
    if [ $euid -ne 0 ]; then
        echo >&2 "Must be root to ${cmd} channel(s)"
        exit 1
    fi
fi

check_prerequisites

case "$cmd" in
    'start')
        if [ -n "${USE_DWAGS}" ]; then
            start_dwags "${USE_DWAGS}" || return 1
        fi
        if [ -n "${USE_LB}" ]; then
            start_lb || exit 1
        fi
        [ "-" = "${cha_tag}" ] || start_channels "${cha_tag}"
        ;;
    'stop')
        [ "-" = "${cha_tag}" ] || stop_channels "${cha_tag}"
        [ -n "${USE_DWAGS}" ] && stop_dwags
        [ -f "${LB_PIDFILE}" ] && stop_lb
        ;;
    'status')
        [ "-" = "${cha_tag}" ] || get_status "${cha_tag}"
        check_pidfile ${GXPM_PIDFILE:?} gxpm
        status_dwags
        [ -f "${LB_PIDFILE}" ] && status_lb
        ;;
    'shutdown') shut_down ;;
    # 'fmtu') print_fmt_uptime ${2:?} ;;
    *) echo >&2 "Unrecognized mode: $cmd"; exit 2 ;;
esac

# __EOF__

