#!/usr/bin/env bash
# @(#) Generate gigaplus.env based on user responses.

errmsg()
{
  echo >&2 "FATAL-ERROR: $@"; exit 1
}

#
# Variables:
#

APP=$0
SCRIPT=$(basename $0)

hostnm=$(hostname)
: ${MAIL_AGENT:='mailx'}

OSNAME=$(uname -s)
case "${OSNAME:?}" in
    'Linux')    dialog_pkg=dialog ;;
    'FreeBSD')  dialog_pkg=cdialog ;;
     *) errmsg "Unsupported OS: ${OSNAME}" ;;
esac
type dialog 2>/dev/null >&2 || errmsg "${dialog_pkg} (>= 1.3) must be available"


#
# Functions
#

ask_admin_email()
{
    local info_txt="Please specify email for a GigA+ administrator.\
 It will be used to send alarms, notices and other important information \
 regarding system performance.\n\n\
 Type 'none' if you want NO email notifications.\n\n\
 Administrator email:"

    local answer=/tmp/evs.aemail.$$
    local lrc=

    dialog --inputbox "${info_txt}" 16 60 ${GXA_ADMIN_EMAIL} 2>$answer
    lrc=$?
    [ $lrc -eq 0 ] && GXA_ADMIN_EMAIL=$(cat $answer)
    rm -f $answer

    return $lrc
}


parse_streaming_mode()
{
    m_HTTP=; m_HLSg=; m_HLSn=

    case "$GXA_OPMODE" in *http*)  m_HTTP='on' ;; esac
    case "$GXA_OPMODE" in *hls-g*) m_HLSg='on' ;; esac
    case "$GXA_OPMODE" in *hls-N*) m_HLSn='on' ;; esac
}


parse_misc_opt()
{
    o_CUSER=; o_PINCPU=; o_ALERT=; o_NOUPK=

    case "${MISC_OPT}" in *cuser*)   o_CUSER=on  ;; esac
    case "${MISC_OPT}" in *pincpu*)  o_PINCPU=on ;; esac
    case "${MISC_OPT}" in *alerts*)  o_ALERT=on  ;; esac
    case "${MISC_OPT}" in *noupk*)   o_NOUPK=on  ;; esac

    [ -n "${o_PINCPU}" ] || GXA_NOPIN_CPU=1
    [ -n "${o_ALERT}"  ] && GXA_OFFCHA_ALERT=1
    [ -n "${o_NOUPK}"  ] && GXA_NO_UPKEEP=1
}


ask_streaming_methods()
{
    local info_txt="Please indicate what types of streaming you will be doing. \
With HLS you have the option of delivering segments using GigA+ gxng module or NginX (or both). \n\n\
Select methods:"
    local answer=/tmp/evs.streaming.$$
    local lrc=

    parse_streaming_mode
    dialog  --title 'Streaming via:' --checklist "${info_txt}" 15 65 3 \
                http  "HTTP (gxng)"     ${m_HTTP:-off} \
                hls-g "HLS  (gxng)"     ${m_HLSg:-off} \
                hls-N "HLS  (NignX)"    ${m_HLSn:-off} 2>${answer}
    lrc=$?
    if [ $lrc -eq 0 ]; then
        GXA_OPMODE=$(cat $answer)
        parse_streaming_mode
    fi

    rm -f $answer
    return $lrc
}


ask_misc_options()
{
    local info_txt="Check the applicable options"
    local answer=/tmp/evs.misc-opt.$$
    local lrc=

    parse_misc_opt
    dialog --title 'Options:' --checklist "${info_txt}" 15 65 4 \
            cuser   "Run as custom user (NOT gigaplus)" ${o_CUSER:-off}     \
            pincpu  "Pin modules to CPU cores"          ${o_PINCPU:-off}    \
            alerts  "Alert on channels offline"         ${o_ALERT:-off}     \
            noupk   "Do NOT watch/restart modules"      ${o_NOUPK:-off} 2>${answer}
    lrc=$?
    if [ $lrc -eq 0 ]; then
        MISC_OPT=$(cat $answer)
        parse_misc_opt
    fi

    rm -f $answer
    return $lrc
}


ask_gng_partition()
{
    local info_txt="How will gxngs handle HTTP and HLS requests? They can either all \
handle BOTH types, OR use some to handle HTTP, some - HLS.\n\n\
Select which way you want it to be:"
    local answer=/tmp/evs.ngs-handle.$$
    local lrc=

    g_Mixed=; g_Dedic=
    case "${GXA_NG_MODE:?}" in
        Mixed)          g_Mixed='on' ;;
        Dedicated)      g_Dedic='on' ;;
    esac

    dialog --radiolist "${info_txt}" 14 60 2 \
        Mixed     'Support HTTP and HLS on all gxngs' ${g_Mixed:-off}  \
        Dedicated 'Use HTTP-only and HLS-only gxngs'  ${g_Dedic:-off}  \
        2>${answer}
    lrc=$?
    [ $lrc -eq 0 ] && GXA_NG_MODE=$(cat $answer)
    rm -f $answer

    return $lrc
}


eval_gng_counts()
{
    local err=
    if [ "${GXA_NG_COUNT}" -lt 0 -o "${GXA_NG_COUNT}" -gt 64 ]; then
        err="${err}non-specific GXNG count must be 0..64\n"
    fi

    if [ -n "${GXA_NOBUF_NG_COUNT}" ]; then
        if [ "${GXA_NOBUF_NG_COUNT}" -lt 0 -o "${GXA_NOBUF_NG_COUNT}" -gt 64 ]; then
            err="${err}HLS GXNG count must be 0..64\n"
        fi
    fi

    if [ -n "$err" ]; then
        dialog --msgbox "$err" 8 40
        return 1
    fi

    if [ "0" -eq "${GXA_NG_COUNT}" ]; then
        if [ "${GXA_NOBUF_NG_COUNT}" -gt 0 ]; then
            local wrn="If you are running only HLS, make sure your gxng configuration \
reflects that and disables bufd subsystem. Your GXNGs would be then started as regular ones, \
not as HLS-specific.\n"

            GXA_NG_COUNT=${GXA_NOBUF_NG_COUNT}
            dialog --msgbox "$wrn" 8 40
        fi
    fi
}


ask_gng_count()
{
    local info_txt="Select how many gxngs should be run (as per license):"
    local dflt_ngs=${GXA_NG_COUNT:?}

    if [ 'Dedicated' = "${GXA_NG_MODE}" ]; then
        info_txt="Your license covers N gxngs that can be run. Some (or all) of them \
may be generic (handle both HTTP and HLS), some may handle HLS requests exclusively. \
Specify either one number, for non-specific gxngs, OR two - separated by slash (as in 2/1). \
The sum must not exceed the total number afforded by your license.\n\n\
Number of generic and HLS-only gxngs (separated by slash):"
        dflt_ngs="${GXA_NG_COUNT}/${GXA_NOBUF_NG_COUNT}"
    fi

    local answer=/tmp/evs.ngs-count.$$
    local lrc=

    while true; do
        dialog --inputbox "$info_txt" 20 60 "${dflt_ngs}" 2>${answer}; lrc=$?
        if [ $lrc -eq 0 ]; then
            local ngs=$(cat $answer)
            set -- $(echo $ngs | tr '/' ' ')
            GXA_NG_COUNT=$1
            GXA_NOBUF_NG_COUNT=$2

            eval_gng_counts && break
        else
            break
        fi
    done

    rm -f $answer
    return $lrc
}


ask_custom_user()
{
    local text="You can run components under a non-privileged user of your choice \
 (NOT gigaplus).\n\n\
 Please provide the name of the custom user:"
    local answer=/tmp/evs.cuser.$$
    local lrc=

    dialog --inputbox "$text" 20 60 "${GXA_USER}" 2>$answer; lrc=$?
    if [ $lrc -eq 0 ]; then
        GXA_USER=$(cat $answer)
        [ 'gigaplus' = "${GXA_USER}" ] && GXA_USER=
    fi

    rm -f $answer
    return $lrc
}


ask_lbal_yesno()
{
    local text='Will you be doing any load-balancing on this host?'
    local opt='--defaultno'
    local lrc=

    [ -n "${GXA_DWAGS}" -o -n "${GXA_USE_FLB}" ] && opt=

    dialog ${opt} --yesno "${text}" 6 50; lrc=$?
    return $lrc
}


ask_dwg_count()
{
    local info_txt="Please select the number of download agents (dwg) \
that you are planning to run on this host. NB: you should have completed \
setting up required third-party products, such as Redis noSQL DB. \
See gxa-lb-setup(5) for details.\n\n\
Number of dwg(1) instances to run:"
    local answer=/tmp/evs.dwg-count.$$
    local lrc=

    dialog --inputbox "${info_txt}" 20 60 ${GXA_DWAGS} 2>${answer}; lrc=$?
    [ $lrc -eq 0 ] && GXA_DWAGS=$(cat $answer)

    rm -f $answer
    return $lrc
}


ask_flb_yesno()
{
    local text='Will you be running a load-balancing gateway on this host?'
    if dialog --defaultno --yesno "${text}" 6 50; then
        GXA_USE_FLB=1
    fi
}


setup_menu()
{
    local answer=/tmp/evs.smenu.$$
    local lrc=
    local lb="dwg=${GXA_DWAGS} flb=${GXA_USE_FLB}"

    [ -n "${GXA_DWAGS}" -o -n "${GXA_USE_FLB}" ] || lb='none'

    dialog --menu 'Environment setup:' 18 70 16             \
        Email       "Admininstrative email \
[${GXA_ADMIN_EMAIL}]"                                       \
        Streaming   "Streaming methods [${GXA_OPMODE}]"     \
        GxngCount   "gxngs modes and counts \
[${GXA_NG_MODE} ${GXA_NG_COUNT:-0}/${GXA_NOBUF_NG_COUNT:-0}]"     \
        LB          "Load balancing [${lb}]"                \
        User        "Custom user [${GXA_USER:--}]"             \
        MiscOpt     "Options [${MISC_OPT}]"                 \
        Done        'Save & exit'                           \
        Quit        'Quit w/o saving'                       \
        2>${answer}; lrc=$?

    [ $lrc -eq 0 ] && CHOICE=$(cat $answer)

    rm -f $answer
    return $lrc
}


init_env_vars()
{
    : ${GXA_ROOT:=/opt/gxa}
    : ${GXA_CHANNEL_ROOT:=${GXA_ROOT:?}/channel}
    : ${GXA_SPEC_DIR:=${GXA_ROOT:?}/etc/spec}
    : ${GXA_LOG_ARCHIVE:=${GXA_ROOT:?}/log/archive}
    : ${GPX_BIN_DIR:=}

    : ${GXA_ADMIN_EMAIL:=}
    : ${GXA_USER:=}
    : ${GXA_NO_UPKEEP:=}
    : ${GXA_OFFCHA_ALERT:=1}

    : ${GXA_NG_COUNT:=2}
    : ${GXA_NOBUF_NG_COUNT:=0}

    : ${GXA_DWAGS:=}
    : ${GXA_USE_FLB:=}
    : ${GXA_NOPIN_CPU:=}
    : ${GXA_CHA_PAUSE:=}

    : ${GXA_OPMODE:='http hls-g'}

    local opt=
    [ -n "${GXA_NOPIN_CPU}" ]       || opt="pincpu $opt"
    [ -n "${GXA_OFFCHA_ALERT}" ]    && opt="alerts $opt"
    [ -n "${GXA_NO_UPKEEP}" ]       && opt="noupk $opt"
    MISC_OPT=$opt

    : ${GXA_NG_MODE:='Mixed'}

    parse_streaming_mode
    parse_misc_opt
}


save_env_profile()
{
    [ "0" -eq "${GXA_NOBUF_NG_COUNT}" ] && GXA_NOBUF_NG_COUNT=
    [ 'none' = "${GXA_ADMIN_EMAIL}" ] && GXA_ADMIN_EMAIL=

cat <<_EOT >${1:?}
# @(#) GigA+ environment settings.
#
# These settings are used by GigA+ scripts
# to initialize runtime environment.
# Most of the settings are can be customized by setup,
# some must be enabled/edited manualy.

# Root directory for GigA+ data (MANDATORY).
GXA_ROOT=${GXA_ROOT:?}

# Root directory for channels (MANDATORY).
GXA_CHANNEL_ROOT=${GXA_CHANNEL_ROOT:?}

# Directory for channel specs.
GXA_SPEC_DIR=${GXA_SPEC_DIR:?}

# Directory for log archives.
# Channel-specific logs reside within channel-named subdirectories,
# others under common" subdirectory.
#
# Re-define when logs are to be stored on a different volume/partition.
#
GXA_LOG_ARCHIVE=${GXA_LOG_ARCHIVE}

# Email for the system administraror (to send system reports to).
GXA_ADMIN_EMAIL=${GXA_ADMIN_EMAIL}

# A hack to allow running debug versions.
# GPX_BIN_DIR=
#

# Define alternate non-privileged user.
GXA_USER=${GXA_USER}

# If set, do NOT launch upkeep (man gxa-setup) for modules.
GXA_NO_UPKEEP=${GXA_NO_UPKEEP}

# Issue an (email) alert if a channel is found OFF (by status check).
GXA_OFFCHA_ALERT=${GXA_OFFCHA_ALERT}

# Number of gxng's to launch: 1..64 (use your own judgement to augment this).
GXA_NG_COUNT=${GXA_NG_COUNT:?}

# Number of ADDITIONAL gxng's NOT using buffers (bufd, man gxng).
# These gxng's could be the ones serving HLS.
#
# NB: if HLS is the ONLY type of content for you, do NOT use this setting,
# simply disable NG buffers in gigaplus.conf.
#
GXA_NOBUF_NG_COUNT=${GXA_NOBUF_NG_COUNT}

# How many downloader agents to use (must configure for LB first, man gxa-lb-setup).
GXA_DWAGS=${GXA_DWAGS}

# Use load balancer (flb) on this host (see man gxa-lb-setup).
GXA_USE_FLB=${GXA_USE_FLB}

# Disable pinning modules to distinct CPU cores.
GXA_NOPIN_CPU=${GXA_NOPIN_CPU}

# Number of seconds to pause between launching channels and checking their status.
# GXA_CHA_PAUSE=${GXA_CHA_PAUSE}

# Protocols utilized. Used by setup (DO NOT EDIT!).
GXA_OPMODE="${GXA_OPMODE}"

# Defines how gxng(1) pool is divided:
#   'Mixed'     = all instances support both HTTP & HLS;
#   'Deicated'  = K instances exclusively support HTTP, L support HLS.
#
# Used by setup (DO NOT EDIT!).
#
GXA_NG_MODE="${GXA_NG_MODE}"

# __EOF__

_EOT
}


confirm_env()
{
    local mode=$1

    local tmpenv=/tmp/evs.tmpenv.$$
    local answer=/tmp/evs.edited.$$
    local lrc=

    CONFIRMED=

    save_env_profile $tmpenv || echo >&2 "Error in save_env_profile"

    if [ 'review' = "${mode}" ]; then
        dialog --title 'Review, edit and confirm' --editbox $tmpenv 40 70 2>${answer}
        [ $? -eq 0 ] || return 1
    fi

    if [ -f "{ENVFILE}" ]; then
        dialog --msgbox "Found [${ENVFILE}]. It will be saved as [${ENVFILE}.orig]" 10 40
        cp ${ENVFILE} ${ENVFILE}.orig
    fi

    mv ${answer} $ENVFILE || return $?
    CONFIRMED=1
    dialog --title 'Done!' --msgbox "GigA+ environment has been saved to [${ENVFILE}]. \
For further alterations, edit it manually or run '${APP} -menu ${ENVFILE}'" 10 60

    rm -f ${answer} ${tmpenv}
    return $lrc
}


full_setup()
{
    ask_admin_email         || return $?
    ask_streaming_methods   || return $?

    ask_gng_partition       || return $?
    ask_gng_count           || return $?

    ask_misc_options        || return $?

    if [ -n "${o_CUSER}" ]; then
        ask_custom_user || return $?
    fi

    if ask_lbal_yesno; then
        ask_dwg_count || return $?
        ask_flb_yesno || return $?
    else
        GXA_DWAGS=; GXA_USE_FLB=
    fi

    confirm_env 'review' || return $?
    QUIT=1

    return 0
}


setup_exit()
{
    local cmd=

    if [ -z "$CONFIRMED" ]; then
        local q='Would you like to edit/confirm the settings?'
        dialog --yesno "$q" 6 50 && cmd='review'
    fi

    confirm_env $cmd || return 1
    QUIT=1
}

#
# Main
#
QUIT=

if [ "$#" -lt 1 ]; then
    echo >&2 "Usage: ${SCRIPT} [-menu] [srcfile] envfile"
    exit 1
fi

if [ '-menu' = "$1" ]; then
    SKIP_FS=1; shift
fi

SRCFILE=${1:?}
ENVFILE=${2}
[ -n "${ENVFILE}" ] || ENVFILE=${SRCFILE}

if [ -f "${SRCFILE}" ]; then
    . ${SRCFILE}
    [ -n "${GXA_ADMIN_EMAIL}" ] && SKIP_FS=1
fi
init_env_vars

if [ -n "${GO_CONFIRM}" ]; then
    confirm_env 'review'; echo rc=$?
    exit 1
fi

[ -n "${SKIP_FS}" ] || full_setup
 while [ -z "${QUIT}" ]; do
    setup_menu || break

    case "$CHOICE" in
        Email)      ask_admin_email         ;;
        Streaming)  ask_streaming_methods   ;;
        GxngCount)  ask_gng_partition && ask_gng_count
                    ;;
        LB)         if ask_lbal_yesno; then
                        ask_dwg_count && ask_flb_yesno
                    else
                        GXA_DWAGS=; GXA_USE_FLB=
                    fi
                    ;;
        User)       ask_custom_user         ;;
        MiscOpt)    ask_misc_options        ;;
        Done)       setup_exit              ;;
        Quit)       break                   ;;
    esac
done

# __EOF__

