#!/usr/bin/env bash
# @(#) Launch script for GigA+ gxws/gxng.
#
# Example: sudo GXA_PFX=/opt/gxa GWS_OPTIONS=-vvv GNG_OPTIONS=-vvv ./gxa-requests start 3
#
#*****************************************************************
# Copyright 2012-2017 Pavel V. Cherenkov (pcherenkov@gigapxy.com)
#
#  This file is part of GigA+.
#
#  GigA+ is commercial software: all rights to it belong
#  to Pavel V. Cherenkov, its author.
#
#*****************************************************************/

errmsg()
{
  echo >&2 $@; exit 1
}

warn()
{
    echo >&2 "WARNING: $@"
}

msg()
{
    echo >&2 $@
}

dbg()
{
    [ -n "${GXA_DEBUG}" ] && echo >&2 "DEBUG: $@"
}



#############################
# Variables
#############################

: ${GXA_PFX:='/opt/gxa'}

RM=/bin/rm
MKDIR=/bin/mkdir
CAT=/bin/cat

OSNAME=$(uname -s)
case "${OSNAME:?}" in
    'Linux'|'FreeBSD') ;;
     *) errmsg "Unsupported OS: ${OSNAME}" ;;
esac


# NB: set GXA_PFX to run in a sandbox.

GPX_PIDFILE_DIR=${GXA_PFX}/run
GPX_LOG_DIR=${GXA_PFX}/log
GPX_BUF_DIR=${GXA_PFX}/buf
GPX_CRASH_DIR=${GXA_PFX}/core

case ${OSNAME} in
    FreeBSD)
        [ -n "${GPX_CONFIG_DIR}" ] || GPX_CONFIG_DIR=/usr/local/etc
        [ -n "${GPX_BIN_DIR}" ] || GPX_BIN_DIR=/usr/local/bin
        ;;
    Linux)
        [ -n "${GPX_CONFIG_DIR}" ] || GPX_CONFIG_DIR=/etc
        [ -n "${GPX_BIN_DIR}" ] || GPX_BIN_DIR=/usr/bin
        ;;
    *) errmsg "${OSNAME} is not supported."
esac

envdir=.
[ -f ${envdir}/gigaplus.env ] || envdir=${GPX_CONFIG_DIR}
envfile="${envdir}/gigaplus.env"

if [ -f "${envfile}" ]; then
    . ${envfile}
else
    errmsg "Missing ${envfile}"
fi

: ${GXA_USER:=gigaplus}

[ '-' = "${GPX_BIN_DIR}" ] && GPX_BIN_DIR=

if [ -n "${GPX_BIN_DIR}" ]; then
    GWS=${GPX_BIN_DIR}/gxws
    GNG=${GPX_BIN_DIR}/gxng
    GPM=${GPX_BIN_DIR}/gxpm
else
    : ${GWS:=gxws}
    : ${GNG:=gxng}
    : ${GPM:=gxpm}
fi

[ -z "${GXA_CONFIG}" ] && GXA_CONFIG=${GPX_CONFIG_DIR}/gigaplus.conf

GXWS_PIDFILE=${GPX_PIDFILE_DIR}/gxws.pid
# The template expands to gng1.pid, gng2.pid, etc.
GXNG_PIDFILE_TPL=${GPX_PIDFILE_DIR}/gxngXX.pid

GXPM_PIDFILE=${GPX_PIDFILE_DIR}/gxpm.pid

GXWS_LOG=${GPX_LOG_DIR}/gxws.log
GXPM_LOG=${GPX_LOG_DIR}/gxpm.log
# The template expands to gng1.log, gng2.log, etc.
GXNG_LOG_TPL=${GPX_LOG_DIR}/gxngXX.log

# Specify any additional command-line options.
GWS_OPTIONS="${GWS_OPTIONS}"
GNG_OPTIONS="${GNG_OPTIONS}"

# Seconds to wait for components to start up.
: ${WAIT_SEC:='1'}


#############################
# Functions
#############################

usage()
{
${CAT} <<__EOM
Usage: $1 OPTIONS start|stop|shutdown|status {num|none} [nobuf-ngs]
OPTIONS:
    --nopm          = do not launch gxpm;
    --killpm        = will stop gxpm (in stop mode);
    --nopin         = no CPU affinity to be set;
    --no-upkeep     = disable upkeep (crash/restart & health monitor);
    --debug mask    = list app(s) in debug mode (ws:ng:pm);
    -t              = trace mode: see commands & trace info;
PARAMETERS:
    num       = number of (gxng) engines to launch [1..64] or 'none'
    nobuf-ngs = additional engines to serve files/segments

Example: $1 --nopm start 3

$1: GigA+ gxpm/gxws/gxng launch script (feel free to modify).
$1 is part of GigA+. GigA+ is Copyright (C) 2016 - 2018 Pavel Cherenkov

__EOM
  exit 1
}


mod_utime()
{
    case "${OSNAME:?}" in
        'Linux')    stat -c '%Y' ${1:?} ;;
        'FreeBSD')  stat -f '%m' ${1:?} ;;
        *) errmsg "Unsupported OS: ${OSNAME}" ;;
    esac
}


lnx_can_suid()
{
    local mode=${1:?}
    local procf='/proc/sys/fs/suid_dumpable'
    case "$mode" in
        --check) cat $procf ;;
        --show) echo "as root: echo 1 >$procf" ;;
        *) errmsg "Invalid mode: $mode" ;;
    esac
}


bsd_can_suid()
{
    local mode=${1:?}
    local setting='kern.sugid_coredump'

    case "$mode" in
        --check) sysctl $setting | cut -d':' -f2 | tr -d ' ' ;;
        --show) echo "as root: sysctl ${setting}=1" ;;
        *) errmsg "Invalid mode: $mode" ;;
    esac
}


get_can_suid()
{
    local mode=${1:?}
    case "${OSNAME:?}" in
        'Linux')    lnx_can_suid ${mode} ;;
        'FreeBSD')  bsd_can_suid ${mode} ;;
    esac
}


get_core_dir()
{
    case "${OSNAME:?}" in
        'Linux')    dirname $(cat /proc/sys/kernel/core_pattern) ;;
        'FreeBSD')  dirname $(sysctl kern.corefile | cut -d':' -f2 | tr -d ' ') ;;
    esac
}


get_fmt_uptime()
{
    local asec=${1:?}

    local dy=$(( ${asec} / 86400 ))
    local dyl=$(( ${asec} % 86400 ))
    local hr=$(( ${dyl} / 3600 ))
    local hrl=$(( ${dyl} % 3600 ))
    local mi=$(( ${hrl} / 60 ))
    local sec=$(( ${hrl} % 60  ))

    printf "%02dd %02dh:%02dm:%02ds" $dy $hr $mi $sec
}


check_prerequisites()
{
  type ${GWS:?} >/dev/null && type ${GNG:?} >/dev/null || \
    errmsg "Missing gxws=[${GWS}], gxng=[${GNG}]"

  type ${GPM:?} >/dev/null || errmsg "Missing gxpm=${GPM}"

  if [ ! -d "${GPX_CONFIG_DIR}" ]; then
    errmsg "$0: config directory [${GPX_CONFIG_DIR}] does not exist."
  fi

  [ '1' = "$(get_can_suid --check)" ] || \
        errmsg "please ensure GigA+ modules CAN core-dump; use $(get_can_suid --show)"

  CORE_DIR=$(get_core_dir)
  if [ '.' = "${CORE_DIR}" ]; then
    warn "cores would go to the current directory, you might want to change that."
    CORE_DIR=
  else
    [ -d "${CORE_DIR}" ] || errmsg "Invalid core directory: ${CORE_DIR}"
  fi

  if [ -z "${SKIP_UPK}" ]; then
    if [ -z "${UPKEEP}" ]; then
        UPKEEP=upkeep
        if ! type ${UPKEEP} >/dev/null 2>&1; then
            [ -n "${GPX_BIN_DIR}" ] && UPKEEP="${GPX_BIN_DIR}/upkeep"
        fi
    fi

    type ${UPKEEP} >/dev/null || UPKEEP=
    if [ -n "${UPKEEP}" -a -z "${GXA_ADMIN_EMAIL}" ]; then
        warn "No GXA_ADMIN_EMAIL, upkeep will be disabled"
    fi

    [ -n "${UPKEEP}" ] || warn "Cannot use upkeep"
    dbg "upkeep: ${UPKEEP}"
  fi


  if [ ! -d "${GPX_LOG_DIR}" ]; then
    ${MKDIR} -p ${GPX_LOG_DIR} || errmsg "$0: failed to mkdir [${GPX_LOG_DIR}]"
  fi
}


upkeep_file()
{
    local cmd_file=${1:?}
    local dir=$(dirname ${cmd_file})
    local base=$(basename ${cmd_file})

    echo ${dir}/upkeep-${base}
}


check_pidfile()
{
    local hush=
    local uptime=

    pid=
    UPK_PID=

    if [ '--hush' = "$1" ]; then
        hush=1; shift
    fi

    local fname=$1
    local app=${2:-'process'}
    local run_msg=${3:-'running'}

    if [ -f "${fname}" ]; then
        uptime=$(($(date +'%s') - $(mod_utime ${fname})))
        pid=$(cat ${fname})
        [ -n "$pid" ] || return 1

        if ! ps -p ${pid} 2>/dev/null >&2; then
            [ -z "$hush" ] && echo "Stale pidfile [${fname}] found; removing."
            rm -f ${fname}
        else
            run_msg="${run_msg}; uptime: $(get_fmt_uptime ${uptime})"
        fi
    fi

    if [ ! -f "${fname}" ]; then
        [ -z "$hush" ] && echo "[${app}] is not running."
        return 1
    fi

    local upk_fname=$(upkeep_file $fname)
    dbg "${app} UPK: ${upk_fname}"

    if [ -f "${upk_fname}" ]; then
        UPK_PID=$(cat ${upk_fname})

        ps -p ${UPK_PID} 2>/dev/null >&2 || UPK_PID=
        if [ -n "${UPK_PID}" ]; then
            run_msg="${run_msg} | U[${UPK_PID}]"
        fi
    else
        dbg "No upkeep pid: ${upk_fname}"
    fi

    [ -z "$hush" ] && echo "${app} [${pid}] ${run_msg}"
    return 0
}


kill_process()
{
    local name=${1:?}
    local cmd_pid=${2:?}
    local upk_pid=$3

    if [ -n "${upk_pid}" ] && ps -p ${upk_pid} 2>/dev/null >&2; then
        dbg "notifying $name U[${upk_pid}]"
        if kill ${upk_pid}; then
            echo "${name}[${cmd_pid}] U[${upk_pid}] asked to stop"
            sleep 0.5
        else
            echo "${name}[${cmd_pid}] U[${upk_pid}] FAILED to stop"
        fi
    fi

    if ps -p ${cmd_pid} >/dev/null 2>&1; then
        dbg "${name}[${cmd_pid}] running (yet)"
        if kill ${cmd_pid}; then
            echo "${name} [${cmd_pid}] stopped"
        else
            echo "${name} [${cmd_pid}] FAILED to stop"
        fi
    else
        dbg "${name}[${cmd_pid}] down (already)"
    fi
}


getoparam()
{
    local flag=${1:?}
    shift; while [ -n "$1" -a "$flag" != "$1" ]; do shift; done
    [ -n "$1" ] && echo $2
}


start_process()
{
    local cmd=${1:?}
    shift
    local options=$@
    local cmd_pidfile=$(getoparam -p $options)
    local lrc=

    [ -n "${cmd_pidfile}" ] || errmsg "-p pidfile must be defined"

    dbg "${cmd} ${options}"
    ${cmd} ${options}; lrc=$?
    [ '0' -eq "$lrc" ] || return $lrc

    if [ -n "${UPKEEP}" ]; then
        local cmd_logfile=$(getoparam -l $options)

        local upk_logfile=$(upkeep_file $cmd_logfile)
        local upk_pidfile=$(upkeep_file $cmd_pidfile)

        local uopt="-d -L $(dirname ${cmd_logfile}) -A ${GPX_CRASH_DIR} -p ${cmd_pidfile} -l ${upk_logfile}"
        uopt="${uopt} -F ${upk_pidfile} -U ${GXA_USER:?}"
        [ -n "$CORE_DIR" ] && uopt="${uopt} -c ${CORE_DIR}"

        sleep 0.3
        check_pidfile --hush ${cmd_pidfile} $(basename ${cmd}) || return 1

        dbg "${UPKEEP} $uopt -- ${cmd} ${options}"

        ${UPKEEP} $uopt -- ${cmd} ${options} >&2 &
        local upk_pid=$!
        sleep 0.3
        if kill -0 ${upk_pid} 2>/dev/null; then
            dbg "${UPKEEP}($upk_pid) running"
            disown ${upk_pid}
        fi
    fi
}


gxy_start()
{
    check_pidfile ${GXWS_PIDFILE} 'GXWS' && return 1

    GNG_BUF=${1:-'1'}
    GNG_FILE=${2:-'0'}
    GNG_NUM=$(( ${GNG_BUF} + ${GNG_FILE} ))

    if [ -z "${NO_GPM}" ]; then
        if ! check_pidfile ${GXPM_PIDFILE} 'GXPM'; then
            echo "Starting GXPM ..."
            start_process ${GPM} -q ${GXPM_OPTIONS} -l ${GXPM_LOG}  -p ${GXPM_PIDFILE} || \
                errmsg "Failed to start GXPM"
        fi
    fi

    echo "Starting GXWS ..."
    start_process ${GWS} -q ${GWS_OPTIONS} -l ${GXWS_LOG} -p ${GXWS_PIDFILE}

    i=1
    while [ "$i" -le "${GNG_NUM:?}" ]; do
        pidfile=`echo ${GXNG_PIDFILE_TPL} | sed "s/XX/$i/g"`
        logfile=`echo ${GXNG_LOG_TPL} | sed "s/XX/$i/g"`

        cpu_affinity=
        if [ -z "${NOPIN}" ]; then
            cpunum=`expr $i - 1`
            cpu_affinity="-P ${cpunum}"
        fi

        gmode='regular'
        nobuf=
        if [ "$i" -gt "${GNG_BUF}" ]; then
            nobuf="--nobufd"
            gmode="file"
        fi

        echo "Starting (${gmode}) GXNG${i} ..."
        start_process ${GNG} -q ${GNG_OPTIONS} -l ${logfile} -p ${pidfile} ${cpu_affinity} ${nobuf}

        i=`expr $i + 1`
    done

    echo "Pausing for ${WAIT_SEC} second(s)."
    sleep ${WAIT_SEC}

    gws_pid=`${CAT} ${GXWS_PIDFILE} 2>/dev/null`
    check_pidfile ${GXWS_PIDFILE} 'GXWS' STARTED || errmsg "See ${GXWS_LOG} for details"

    ng_err=0; i=1
    while [ "$i" -le "${GNG_NUM}" ]; do
        pidfile=`echo ${GXNG_PIDFILE_TPL} | sed "s/XX/$i/g"`
        logfile=`echo ${GXNG_LOG_TPL} | sed "s/XX/$i/g"`

        if ! check_pidfile ${pidfile} 'GXNG' STARTED; then
            let ng_err+=1
            msg "GXNG${i} failed to start; see ${logfile} for details"
        fi

        i=`expr $i + 1`
    done

    check_pidfile ${GXPM_PIDFILE} 'GXPM' STARTED || msg "See ${GXPM_LOG} for details"

    if [ "$ng_err" -gt 0 ]; then
        echo "$0: some GNG's could not start, shutting down GWS."
        kill ${gws_pid}
    fi
}


gxy_status()
{
    [ -z "${NO_GPM}" ] && check_pidfile ${GXPM_PIDFILE} 'GXPM'
    check_pidfile ${GXWS_PIDFILE} 'GXWS' || return 1

    ls ${GPX_PIDFILE_DIR}/gxng*.pid 2>/dev/null | while read fname; do
        check_pidfile ${fname} 'GXNG'
    done
}


gxy_stop()
{
    if [ -n "${KILL_GPM}" ]; then
        if check_pidfile ${GXPM_PIDFILE} 'GXPM'; then
            kill_process 'GXPM' ${pid} ${UPK_PID}
        fi
    fi


    check_pidfile ${GXWS_PIDFILE} 'GXWS' || return 1
    kill_process 'GXWS' ${pid} ${UPK_PID}

    # if ws.ng.force_shutdown is enabled, GNG's should quit.
    sleep ${WAIT_SEC}

    local i=1
    ls ${GPX_PIDFILE_DIR}/gxng*.pid 2>/dev/null | while read fname; do
        check_pidfile ${fname} "GXNG${i}" || return 1
        kill_process "GXNG{i}" ${pid} ${UPK_PID}
        let i+=1
    done
}


#############################
# Main
#############################

[ $# -lt 1 ] && usage $0

while true; do
    if [ '--nopm' = "$1" ]; then
        NO_GPM=1; shift
        continue
    fi
    if [ '--killpm' = "$1" ]; then
        KILL_GPM=1; shift
        continue
    fi
    if [ '--nopin' = "$1" ]; then
        NOPIN=1; shift
        continue
    fi

    if [ '--no-upkeep' = "$1" ]; then
        SKIP_UPK=1
        shift; continue
    fi

    if [ '-t' = "$1" ]; then
        GXA_DEBUG=1
        shift; continue
    fi

    if [ "--debug" = "$1" ]; then
        dmask=${2}
        case "$dmask" in
            *ws*)   GWS_OPTIONS="-vvv ${GWS_OPTIONS}";
                    echo >&2 "[gxws-debug]"
                    ;;
        esac
        case "$dmask" in
            *ng*)  GNG_OPTIONS="-vvv ${GNG_OPTIONS}"
                    echo >&2 "[gxng-debug]"
                    ;;
        esac
        case "$dmask" in
            *pm*)   GXPM_OPTIONS="-L debug ${GXPM_OPTIONS}"
                    echo "[gxpm-debug]"
                    ;;
        esac
        shift 2; continue
    fi
    break
done

check_prerequisites

case "$1" in
    'start'|'START')
        if [ $# -ge 2 ]; then
            if [ "none" = "$2" ]; then
                ngnum=0; nobuf_ngs=0
            else
                # Number of (gng) engines to launch.
                ngnum=`expr $2 + 0 2>/dev/null`
                [ "0" != "$2" -a $? -ne 0 ] && errmsg "$0: invalid parameter format."

                if [ $ngnum -lt 0 -o ${ngnum} -gt 64 ]; then
                    errmsg "num [${ngnum}] is out of range."
                fi

                nobuf_ngs=0
                [ -n "$3" ] && nobuf_ngs=`expr "$3 + 0" 2>/dev/null`
            fi
        else
            errmsg "Missing mandatory parameter num = number of engines"
        fi
        gxy_start ${ngnum} ${nobuf_ngs}
        ;;

    'stop'|'STOP')
        gxy_stop
        ;;
    'shutdown'|'SHUTDOWN')
        KILL_GPM=1; gxy_stop
        ;;

    'status'|'STATUS')
        gxy_status
        ;;
    *) errmsg "Unrecognized command: $1" ;;
esac


# __EOF__

