-- Get next load-balanced URL using metric-based balancing
--
-- IN:  KEYS[1] = segment KEY (Example: 'v1/20170504-03/33303992.ts');
-- IN:  ARGV[1] = metric name (Example: 'cpu_idle');
-- IN:  ARGV[2] = use min|max criteria;
-- IN:  ARGV[3] = options:
--              '-S' = error if sensor not found;
-- OUT: load-balanced URL for the segment.
--

local function get_sensor_value(url, stype)
    local i,j = string.find(url, "http://.+[0-9]+/")
    if not i then
        return nil, "#err: no host key for: " .. url
    end

    local hostk = string.sub(url, i, j)

    local alias = redis.call("HGET", "host_alias", hostk)
    if not alias then
        return nil, "#err: no host alias for " .. hostk
    end

    local sensor = "S" .. ":" .. stype .. ":" .. alias
    local val = redis.call("GET", sensor)
    if not val then
        return nil, "#err: no value for " .. sensor
    end

    return tonumber(val)
end

--
-- Main
--

local locK = "LOC:" .. KEYS[1]
local posK = "POS:" .. KEYS[1]

local sensor_type = ARGV[1]
local criteria = string.lower(ARGV[2])

local flag_err_no_sensor = nil

local options = ARGV[3]
if "-" == options then options = nil end

if options then
-- '-S' = error if sensor not found;
    flag_err_no_sensor = string.find(options, "-S")
end

if ("min" ~= criteria) and ("max" ~= criteria) then
    return redis.error_reply("#err: wrong criteria=" .. criteria)
end

local num_urls = redis.call("ZCARD", locK)
if num_urls < 1 then
    return redis.error_reply("#err: empty " .. locK)
end

local best_val = nil
local found = false

-- Find the best sensor value (by given criteria).
--
local Urls = redis.call("ZRANGE", locK, 0, -1)
for i = 1, num_urls do
    local metric, errmsg = get_sensor_value(Urls[i], sensor_type)
    if metric then
        if not best_val then
            best_val = metric
        end

        if "max" == criteria then
            if metric > best_val then
                best_val = metric
            end
        end

        if "min" == criteria then
            if metric < best_val then
                best_val = metric
            end
        end
    else
        if flag_err_no_sensor then
            return redis.error_reply("#err: absent (sensor): " .. tostring(errmsg))
        end
        -- Unless flag is set, when a sensor is absent (server down?),
        -- skip the server.
    end
end

-- DEBUG ONLY, uncomment if in trouble.
--
--if best_val >= 0 then
--    return redis.error_reply("#dbg: b=" .. tostring(best_val))
--end


if not best_val then
-- No value picked: no sensors were available (no feed?, servers down?).
   return redis.error_reply("#err: no sensor-based values")
end

-- Advance current position:
-- let other URLs with the same value participate.
--
local pos = redis.call("INCR", posK)
if (pos >= num_urls) or (pos < 0) then pos = 0 end
local p = tonumber(pos)

-- Find the next URL matching the best sensor value.
local count = 0
repeat
    local url = Urls[p + 1]
    if not url then
        return redis.error_reply("#err: nil URL at pos=" .. tostring(p + 1))
    end

    local metric, errmsg = get_sensor_value(url, sensor_type)
    if metric then
        found = metric == best_val
        if found then
            break
        end
    else
        if flag_err_no_sensor then
            return redis.error_reply("#err: (Url) absent (sensor): " .. tostring(errmsg))
        end
    end

    p = p + 1
    if p >= num_urls then p = 0 end

    count = count + 1
until count >= num_urls

if found then
    pos = p
else
    return redis.error_reply("#err: cannot find best value=" .. best_val);
end

redis.call("SET", posK, pos)

return redis.call("ZRANGE", locK, pos, pos)

-- EOF --

